;; -*- Lisp -*-

;;;; Rule package for Trivial Rule Engine using JTMS
;;;; Version 8, Last edited 3/19/1991

;;; Copyright (c) 1989, 1990, 1991, Kenneth D. Forbus, Northwestern University,
;;; Johan de Kleer and Xerox Corporation.
;;; All rights reserved.

(proclaim '(special *JTRE* *end-forms* *rule-indexing* *bound-vars*))

;;; The next three variable are used during rule construction.

(defvar *bound-vars* nil)    ;; Keeps track of lexical environment
(defvar *end-forms* nil)     ;; Caches procedure definitions
(defvar *rule-indexing* nil) ;; Caches rule indexing forms

;; These help ensure unique names for internally generated procedures.
;; When writing a system with several files of rules, place 
;; (RULE-PREFIX <your string>) in front of the file. 
(defvar *file-counter* 0)  
(defvar *file-prefix* "")

(defstruct (rule :conc-name
		 (:print-function jtre-rule-printer))
  counter      ; Unique ID for easy lookup
  jtre         ; The JTRE it is part of
  class        ; Class of associated pattern
  matcher      ; Procedure that performs the match.
  body)        ; Procedure that does the work.

(defun jtre-rule-printer (r st ignore)
  (declare (ignore ignore))
  (format st "<Rule ~D>" (rule-counter r)))

(defmacro Rule-File (prefix)
  `(eval-when (compile load eval)
     (setq *file-counter* 0)
     (setq *file-prefix* ,prefix)))

;;;; Building and installing rules

(defmacro rule (triggers &rest body) (do-rule triggers body))

(defun do-rule (triggers body)
  (let ((*end-forms* nil) ;; accumulates procedure definitions
	(*rule-indexing* nil) ;; accumulates indexing forms
	(*JTRE* *JTRE*))    ;; the JTRE rule will use.
    ;; Uncomment these if you have a serious rule expansion problem to track down.
    ;;(format t "~% Adding rule with triggers: ") (pprint triggers)
    ;;(format t "~% And body: ") (pprint body)
  (add-rule (car triggers)
	    (subst 'internal-rule
		   'rule
		   (make-nested-rule
		     (cdr triggers) body)))
  `(progn ,@ *end-forms* ,@ *rule-indexing*)))

(defmacro internal-rule (triggers &rest body)
  `(add-internal-rule ,(car triggers)
     ,(make-nested-rule (cdr triggers) body)))

(defun make-nested-rule (triggers body)
  (cond ((null triggers) body)
	(t `((add-internal-rule ,(car triggers)
	       ,(make-nested-rule (cdr triggers) body))))))

(defun add-rule (trigger body)
  (push (build-rule trigger body) *rule-indexing*)
  nil)

(defmacro add-internal-rule (trigger body)
  (build-rule trigger body))

;;;; Building a rule involves analyzing the trigger to see what
;;   special-purpose unifier is needed, constructing a function
;;   to do the work of the body, and building a form to index it.

(defun build-rule (trigger body &aux match-procedure body-procedure)
  (multiple-value-bind (pattern condition var test)
		       (parse-rule-trigger trigger)
   (setq match-procedure (generate-match-procedure pattern var test condition))   
   (setq body-procedure (generate-body-procedure pattern condition var body))
   (push match-procedure *end-forms*)
   (push body-procedure *end-forms*)
   `(insert-rule
     (get-class ,(get-trigger-class pattern))
     ;return form to index rule
     (function ;the match function for rule
       ,(if *bound-vars*
	    `(lambda (p)
	       (,(cadr match-procedure) p ,@ *bound-vars*))
	  (cadr match-procedure)))
     (function ;;the body function for rule
       ,(if (or *bound-vars*
		(not (eq condition ':INTERN)))
	    (let ((tv (nreverse
			(pattern-free-variables trigger))))
	      (unless (eq condition ':INTERN) (push 'TRIGGER-NODE tv))
	      `(lambda ,tv
		 (,(cadr body-procedure) ,@ tv
		  ;(fn-name parameters)
		  ,@ (scratchout tv *bound-vars*))))
	      (cadr body-procedure))))))

(defun parse-rule-trigger (trigger)
  ;; Trigger syntax is now (<condition> <pattern> .  <options>)
  ;; where <condition> =  :INTERN | :IN | :OUT
  ;;  and <options> can be empty, or 
  ;;  :TEST <code> and/or :VAR <var>, where <code> must be
  ;;  non-nil for the match to succeed, and <var> will be
  ;;  bound to the whole pattern.
  (values (cadr trigger)
	  (cond ((member (car trigger) '(:INTERN :IN :OUT)) (car trigger))
		(t (error "~% Unknown belief condition ~A in trigger ~A."
			  (car trigger) trigger)))
	  (cadr (member ':VAR (cddr trigger)))
	  (cadr (member ':TEST (cddr trigger)))))

(defun get-trigger-class (trigger)
  (cond ((variable? trigger)
	 (if (member trigger *bound-vars*)  trigger
	     (error "~%Trigger class is unbound -- ~A."
		    trigger)))
	((atom trigger)  (list 'QUOTE trigger))
	(t (get-trigger-class (car trigger)))))

;;;; Generating the body function

(defmacro with-pushed-variable-bindings (new-bindings
					  &rest body)
  `(let ((*bound-vars* (append ,new-bindings
			       (scratchout ,new-bindings
					   *bound-vars*))))
     ,@ body))

(defun generate-body-procedure (pattern condition var body
					&aux newly-bound env fname)
  (setq newly-bound (pattern-free-variables pattern))
  (if var (push var newly-bound))
  (setq body (with-pushed-variable-bindings
	       newly-bound (fully-expand-body body)))
  (setq env (append newly-bound
		    (scratchout newly-bound *bound-vars*)))
  (unless (eq condition ':INTERN) (push 'trigger-node env))
  (setq fname (generate-rule-procedure-name pattern))
  `(defun ,fname ,env
     ,@ (cond ((eq condition ':INTERN) body) ;; Just do it
	      (t ;; Must check and see if the node's belief state
	         ;; matches the rule's requirements
	       `((cond ((,(case condition
				(:IN 'in-node?)(:OUT 'out-node?)
				(t (error "~A bad condition -- GBF"
						 condition)))
			 TRIGGER-NODE) ,@ body)
		       (t (push (list ',fname ,@ env)
				,(ecase condition
				       (:IN `(tms-node-in-rules TRIGGER-NODE))
				       (:OUT `(tms-node-out-rules TRIGGER-NODE)
					     ))))))))))

(defun scratchout (l1 l2)  ;non-destructive and order-preserving
  (dolist (el1 l1 l2) (setq l2 (remove el1 l2))))

(defun generate-rule-procedure-name (pattern)
  (intern (format nil "~A-~A-~A" *file-prefix* pattern (incf *file-counter*))))

;;;; Recursive macroexpansion
;;   Knowing what forms to skip over is still implementation
;;   dependent.  It gets hairier when one allows assertions to
;;   have procedural side-effects.

(defun fully-expand-body (body)
  (cond ((null body) nil)
	((not (listp body)) body)
	((symbolp (car body))
	 (case (car body)
	   (LAMBDA `(lambda ,(cadr body)
		      ,@ (fully-expand-body (cddr body))))
	   (DO `(do ,(cadr body) ,(caddr body)
		  ,@ (cdddr body)))
	   #+Lucid
	   (BQ-LIST* `(BQ-LIST* ,@ (mapcar #'fully-expand-body (cdr body))))
	   ((DOTIMES DOLIST LET PROG PROGV)
	    `(,(car body) ,(cadr body)
	      ,@ (fully-expand-body (cddr body))))
	   (AND (cond ((cdr body) `(AND ,@ (fully-expand-body (cdr body))))
		      (t body)))
	   (OR (cond ((cdr body) `(OR ,@ (fully-expand-body (cdr body))))
		     (t body)))
	   ((PROGN BLOCK)
	    `(,(car body) ,@ (mapcar #'fully-expand-body (cdr body))))
	   (MULTIPLE-VALUE-BIND `(multiple-value-bind ,(cadr body) ,(fully-expand-body (caddr body)) 
				   ,@ (fully-expand-body (cdddr body))))	   
	   (t (let ((new-body (macroexpand body)))
		(cons (fully-expand-body (car new-body))
		      (fully-expand-body (cdr new-body)))))))
	(t (let ((new-body (macroexpand body)))
	     (cons (fully-expand-body (car new-body))
		   (fully-expand-body (cdr new-body)))))))

;;;; Running rules

(defun insert-rule (class matcher body &aux rule)
  (let ((*JTRE* (class-jtre class)))
    (setq rule (make-rule :matcher matcher
			  :body body
			  :class class
			  :counter (incf (jtre-rule-counter *JTRE*))))
    (push rule (class-rules class))
    (dolist (candidate (class-facts class))
	    (try-rule-on rule candidate))))

(defun try-rules (datum)
  (dolist (rule (class-rules (datum-class datum)))
    (try-rule-on rule datum)))

(defun try-rule-on (rule datum)
  (let ((*JTRE* (datum-jtre datum)))
    (multiple-value-bind (okay? bindings node?)
			 (funcall (rule-matcher rule)
				  (datum-lisp-form datum))
			 (when okay?
			       (when node?
				     (push (datum-tms-node datum) bindings))
			       (enqueue (cons (rule-body rule) bindings) *JTRE*)))))

(defun run-rules (&optional (*JTRE* *JTRE*))
  (do ((form (dequeue *JTRE*) (dequeue *JTRE*))
       (counter 0 (1+ counter)))
      ((null form)
       (debugging-jtre "~%    ~A rules run."  counter)
       (incf (jtre-rules-run *JTRE*) counter))
    (apply (car form) (cdr form))))

(defun rules-waiting? (jtre) (jtre-queue jtre))

(defun enqueue (new j) (push new (jtre-queue j)))

(defun dequeue (jtre) (pop (jtre-queue jtre)))

;;;; Display routines

(defun show-rules (&optional (*JTRE* *JTRE*) (stream *standard-output*))
  (format t "~%There are ~D rules in ~A:" 
	  (jtre-rule-counter *JTRE*) (jtre-title *JTRE*))
  (format stream "~% ~A queued." (if (null (jtre-queue *JTRE*)) "None"
				   (length (jtre-queue *JTRE*))))
  (map-class #'(lambda (class)
		 (dolist (rule (class-rules class))
			 (print-rule rule stream)))))

(defun print-rule (rule &optional (stream *standard-output*))
  (format stream "~% ~A: ~A, ~A" rule
	  (rule-matcher rule) (rule-body rule)))

(defun test-rule-expansion ()
 (pprint (macroexpand
	  '(rule ((:in (implies ?p ?q) :var ?f1)
		  (:in ?p)) (rassert! ?q (CE ?f1 ?p))))))

(defun get-rule (num &optional (*JTRE* *JTRE*))
  (map-class #'(lambda (class)
		 (dolist (rule (class-rules class))
			 (when (= (rule-counter rule) num)
			       (return-from GET-RULE rule))))))
